#include "PeriodicBC.h"
#include "BasisFunction.h"
#include "DOFVector.h"
#include "LeafData.h"
#include "DOFMatrix.h"
#include "Traverse.h"
#include "Boundary.h"
#include "VertexVector.h"

namespace AMDiS {

  std::vector<PeriodicDOFMapping*> PeriodicDOFMapping::mappings_;

  PeriodicDOFMapping* 
  PeriodicDOFMapping::providePeriodicDOFMapping(const BasisFunction *basFcts)
  {
    std::vector<PeriodicDOFMapping*>::iterator it;
    std::vector<PeriodicDOFMapping*>::iterator end = mappings_.end();
    for (it = mappings_.begin(); it != end; ++it)
      if ((*it)->basFcts_ == basFcts)
	return *it;

    PeriodicDOFMapping *newMapping = new PeriodicDOFMapping(basFcts);
    mappings_.push_back(newMapping);
    return newMapping;
  }


  PeriodicDOFMapping::PeriodicDOFMapping(const BasisFunction *basFcts) 
    : basFcts_(basFcts) 
  {
    FUNCNAME("PeriodicDOFMapping::PeriodicDOFMapping()");
    TEST_EXIT(basFcts_->getDim() > 1)("dim == 1\n");  
    int  num = basFcts_->getNumber();
    DimVec<double> *lambda;
    for (int i = 0; i < num; i++) {
      lambda = basFcts_->getCoords(i);
      indexOfCoords_[*lambda] = i;
    }
  }

  PeriodicDOFMapping::~PeriodicDOFMapping()
  {
    std::map<DimVec<int>, DegreeOfFreedom*, DimVecLess<int> >::iterator it;
    for (it = dofPermutation_.begin(); it != dofPermutation_.end(); ++it)
      if (it->second)
	delete [] it->second;
  }

  const DegreeOfFreedom *PeriodicDOFMapping::getDOFPermutation(const DimVec<int> &vertexPermutation)
  {
    FUNCNAME("PeriodicDOFMapping::getDOFPermutation()");

    if(dofPermutation_[vertexPermutation] == NULL) {
      //MSG("new dof permutation needed\n");

      int i, j;
      int dim = basFcts_->getDim();
      int num = basFcts_->getNumber();

      int sum = 0;
      for(i = 0; i < dim + 1; i++) {
	sum += i - vertexPermutation[i];
	TEST_EXIT(vertexPermutation[i] < dim + 1)
	  ("invalid vertexPermuation\n");
      }
      TEST_EXIT(sum == 0)("invalid vertexPermutation\n");

      // create dof permutation
      DimVec<double> *lambda;
      DimVec<double> newLambda(dim, NO_INIT);

      DegreeOfFreedom *mapping = new DegreeOfFreedom[num]; 

      for(i = 0; i < num; i++) {
	lambda = basFcts_->getCoords(i);
	for(j = 0; j < dim + 1; j++) {
	  newLambda[vertexPermutation[j]] = (*lambda)[j];
	}
	mapping[i] = indexOfCoords_[newLambda];
      }

      dofPermutation_[vertexPermutation] = mapping;
    }

    return dofPermutation_[vertexPermutation];
  }


  PeriodicBC::PeriodicBC(BoundaryType type, FiniteElemSpace *rowFESpace_) 
    : BoundaryCondition(type, rowFESpace_, NULL),
      masterMatrix_(NULL)
  {
    if(rowFESpace->getMesh()->getDim() > 1) {
      periodicDOFMapping_ = 
	PeriodicDOFMapping::providePeriodicDOFMapping(rowFESpace_->getBasisFcts());
    } else {
      periodicDOFMapping_ = NULL;
    }
  }

  PeriodicBC::~PeriodicBC()
  {}

  void PeriodicBC::initMatrix(DOFMatrix* matrix)
  {
    FUNCNAME("PeriodicBC::initMatrix()");

    if (!masterMatrix_) {
      masterMatrix_ = matrix;

      Mesh *mesh = matrix->getRowFESpace()->getMesh();

      associated_ = mesh->getPeriodicAssociations()[boundaryType];
      
      TEST_EXIT(associated_)("no associations for periodic boundary condition %d\n",
			     boundaryType);

      const BasisFunction *basFcts = rowFESpace->getBasisFcts();
      neighIndices_ = new DegreeOfFreedom[basFcts->getNumber()];
    }
  }

  void PeriodicBC::fillBoundaryCondition(DOFMatrix *matrix,
					 ElInfo *elInfo,
					 const DegreeOfFreedom *dofIndices,
					 const BoundaryType *localBound,
					 int nBasFcts)
  {
    if (matrix == masterMatrix_) {
  
      int dim = rowFESpace->getMesh()->getDim();
      if (dim > 1) {
	int i, j;
	DOFAdmin *admin = rowFESpace->getAdmin();

	FixVec<int, WORLD> elFace(dim, NO_INIT);
	FixVec<int, WORLD> neighFace(dim, NO_INIT);
	DimVec<int> vertexPermutation(dim, NO_INIT);

	const BasisFunction *basFcts = rowFESpace->getBasisFcts();
	int num = basFcts->getNumber();

	Element *element = elInfo->getElement();

	DimVec<DegreeOfFreedom> periodicDOFs(dim-1, NO_INIT);

	int vertex, index, side;

	GeoIndex sideGeoIndex = INDEX_OF_DIM(dim-1, dim);

	for (side = 0; side < dim + 1; side++) {

	  if (elInfo->getBoundary(sideGeoIndex, side) == boundaryType) {

	    for (vertex = 0; vertex < dim; vertex++) {
	      index = element->getVertexOfPosition(sideGeoIndex,
						   side,
						   vertex);
	      periodicDOFs[vertex] = (*associated_)[dofIndices[index]];
	    }

	    Element *neigh = elInfo->getNeighbour(side);
	    basFcts->getLocalIndices(neigh, admin, neighIndices_);

	    int oppVertex = 0;
	    for(i = 0; i < dim + 1; i++) {
	      // get vertex permutation
	      if(i == side) {
		vertexPermutation[i] = 0;
	      } else {
		DegreeOfFreedom periodicDOF = 
		  periodicDOFs[element->getPositionOfVertex(side, i)];

		for(j = 0; j < dim + 1; j++) {
		  if(neigh->getDOF(j, 0) == periodicDOF) break;
		}
		vertexPermutation[i] = j;
	      }
	      oppVertex += i - vertexPermutation[i];
	    }
	    vertexPermutation[side] = oppVertex;

	    // get DOF permutation
	    const DegreeOfFreedom *dofPermutation = 
	      periodicDOFMapping_->getDOFPermutation(vertexPermutation);
	
	    // set associated dofs
	    for(i = 0; i < num; i++) {
	      if((*(basFcts->getCoords(i)))[side] == 0) {
		(*associated_)[dofIndices[i]] = neighIndices_[dofPermutation[i]];
	      }
	    }
	  }
	}
      }
    }
  }

  void PeriodicBC::exitMatrix(DOFMatrix* matrix)
  {
    FUNCNAME("PeriodicBC::exitMatrix()");

    ERROR_EXIT("Not yet implemented, if you have an idea how to do it, than make it!!!\n");

    TEST_EXIT(matrix)("no matrix\n");

    if (matrix == masterMatrix_) {
      delete [] neighIndices_;
      masterMatrix_ = NULL;
    }

    using namespace mtl;

    associated_->print();

    // Compute reorder matrix (newRow and newCol yields transposed!!!)
    matrix::traits::reorder<>::type       R= matrix::reorder(*associated_);
    DOFMatrix::base_matrix_type           &A= matrix->getBaseMatrix(), B, D, E, TR;

    A*= 0.5;
    // Half of entries with decreased row index + half of the strict lower origing
    B= strict_upper(R) * A + strict_lower(A);
    // Add half of entries with same row and decreased column index + half of increased columns
    // Remark regarding column permutation: trans(strict_upper(trans(X))) == strict_lower(X)
    // B+= bands(trans(R), 0, 1) * (A * strict_lower(R) + lower(A));  -> make this work one day!!!
    TR= trans(R);
    D= bands(TR, 0, 1);
    E= A * strict_lower(R) + lower(A);
    B+= D * E;
    swap(A, B);   
  }

  void PeriodicBC::exitVector(DOFVectorBase<double>* vector)
  {
    FUNCNAME("PeriodicBC::exitVector()");

    double entry;

    DegreeOfFreedom dof;
    DegreeOfFreedom newDOF;

    DOFIterator<double> vecIt(vector, USED_DOFS);

    Mesh *mesh = vector->getFESpace()->getMesh();
    VertexVector *associated_ = mesh->getPeriodicAssociations()[boundaryType];

    for (vecIt.reset(); !vecIt.end(); ++vecIt) {
      dof = vecIt.getDOFIndex();

      newDOF = (*associated_)[dof];

      if (dof < newDOF) {
	entry = ((*vector)[dof] + (*vector)[newDOF]) * 0.5;
	(*vector)[dof] = entry;
	(*vector)[newDOF] = entry;
      }
    }
  }

}
